/* --------------------------- sbio.c ---------------------------------	*/
/* Support programs for SoundBlaster access.				*/
/*									*/
/* Written by Eyal Lebedinsky [eyal@ise.canberra.edu.au]		*/
/*									*/
/* --------------------------------------------------------------------	*/

#include <stdio.h>
#include <conio.h>

#include "sbio.h"

#define rangeof(x)	(sizeof(x)/sizeof(x[0]))

#define DSP		dsp_port

/* Port Addresses
*/
#define DSPRESET	(DSP+0x6)	/* reset (write only) */
#define DSPREAD		(DSP+0xA)	/* read data (read-only) */
#define DSPWRITE	(DSP+0xC)	/* write data (write-only) */
#define DSPRSTAT	(DSP+0xE)	/* read buffer status (read-only) */
#define DSPWSTAT	(DSP+0xC)	/* write buffer status (read-only) */

/* Status Port Bits
*/
#define NOWRITE		0x080		/* write buffer busy */
#define YESREAD		0x080		/* read buffer ready */

/* MPU commands
*/
#define SAMPLEWRITE	0x010		/* write audio sample to DSP */
#define SAMPLEREAD	0x020		/* read audio sample from DSP */
#define MIDIREADD	0x030		/* read (direct) */
#define MIDIUART	0x034		/* UART (polled) */
#define MIDIWRITE	0x038		/* write (direct) */
#define SETRATE		0x040		/* set sampling rate */
#define ASKVERSION	0x0E1		/* get version */

#define dsp_read()	inp (DSPREAD)
#define dsp_write(b)	outp (DSPWRITE, b)
#define dsp_rstat()	inp (DSPRSTAT)
#define dsp_wstat()	inp (DSPWSTAT)
#define input_ready()	(dsp_rstat () & YESREAD)
#define output_ready()	(!(dsp_wstat () & NOWRITE))

static int		dsp_port = 0x220;
static unsigned char	inbuf[4096];
static int		bufhead = 0, buftail = 0, bufsize = 0;

#define LEN		sizeof(inbuf)
#define END		(LEN - 1)
#define INC(x)		(x = (x >= END) ? 0 : x+1)

static int	dspmode = 0;
static int	uartmode = 0;

static int
buf_put (c)
int	c;
{
	int	r;

	if (LEN == bufsize) {
		r = -1;
	} else {
		r = 0;
		inbuf[buftail] = (unsigned char)c;
		INC (buftail);
		++bufsize;
	}
	return (r);
}

static void
get_in ()
{
	long	timeout;

	for (timeout = 30000L; input_ready () && timeout > 0; timeout--)
		buf_put (dsp_read ());
}

static int
buf_get ()
{
	int	c;

	get_in ();

	if (bufsize) {
		c = inbuf[bufhead];
		INC (bufhead);
		--bufsize;
	} else
		c = -1;
	return (c);
}

static int
wait_out ()
{
	long	timeout;

	for (timeout = 30000L; !output_ready (); timeout--) {
		if (timeout <= 0)
			return (-1);
	}
	return (0);
}

static int
dsp_out (c)
int	c;
{
	if (wait_out ())
		return (-1);
	dsp_write (c);
	return (0);
}

static int
wait_in ()
{
	long	timeout;

	for (timeout = 30000L; !input_ready (); timeout--) {
		if (timeout <= 0)
			return (-1);
	}
	return (0);
}

static int
dsp_in ()
{
	if (wait_in ())
		return (-1);
	return (dsp_read ());
}

static void
delayit ()
{
	long	timeout;

	for (timeout = 300000L; timeout > 0; timeout--)
		dsp_rstat();
}

static int
get_ack ()
{
	int	timeout, c;

	for (timeout = 10L; timeout > 0; timeout--) {
		if ((c = dsp_in ()) < 0)
			continue;
		if (0x0aa == c)
			return (0);
	}
	return (-2);
}

int
ResetMPU ()
{
	int	n;

	bufhead = buftail = bufsize = 0;
	uartmode = 0;

	for (n = 0;; n++) {
		if (n > 2)
			return (-1);
		outp (DSPRESET, 1);
		delayit ();
		outp (DSPRESET, 0);
		if (!get_ack ())
			break;

	}
	return (0);
}

int
CloseMPU ()
{return (0);}

int
SetUART ()
{
	int	n;

	if (!dsp_out (ASKVERSION)) {
		n = dsp_in ();
		dsp_in ();
	} else
		n = 0;

	if (n >= 2 && !dsp_out (MIDIUART))
		uartmode = 1;

	return (0);
}

int
SendData (c)
int	c;
{
	get_in ();

	if (!uartmode) {
		if (dsp_out (dspmode = MIDIWRITE)) {
#ifdef DEBUG
printf (" -1 ", c);
#endif
			return (-1);
		}
	}

	if (dsp_out (c)) {
#ifdef DEBUG
printf (" -2 \n", c);
#endif
		return (-2);
	}

#ifdef DEBUG
printf (" %02x ", c);
#endif
	return (0);
}

int
ReceiveData ()
{
	int	c;
	long	timeout;

	if (!uartmode) {
		if (dspmode != MIDIREADD) {
			if (dsp_out (dspmode = MIDIREADD)) {
#ifdef DEBUG
printf ("[-1]");
#endif
				return (256);
			}
		}
	}

	for (timeout = 100L;; timeout--) {
		if ((c = buf_get ()) >= 0)
			break;
		if (timeout <= 0) {
#ifdef DEBUG
printf ("[-2]");
#endif
			return (256);
		}
	}
#ifdef DEBUG
printf ("[%02x]", c);
#endif
	return (c);
}

int
SetRate (long rate)
{
	int	period;

	if (dsp_out (SETRATE))
		return (-1);

	if (rate <= 1000000L/256)
		period = 256;
	else
		period = 256 - 1000000L/rate;

	return (dsp_out (period));
}

int
ReadSample ()
{
	if (dsp_out (SAMPLEREAD))
		return (-1);

	return (dsp_in ());
}

int
WriteSample (int i)
{
	if (dsp_out (SAMPLEWRITE))
		return (-1);

	return (dsp_out (i));
}


